/*
 * Copyright 2014 Pierre-Edouard Portier, Mazen Alsarem
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <search.h>
#include <time.h>
#include <unistd.h>
#include "dbg.h"
#include "svdlib.h"
#include "str.h"
#include "ldpr.h"

/* the importance of the teleportation factor is controlled by the alpha 
 * parameter. At each step, the random walker choose to either reach one of 
 * the neighbour positions in the graph with probability alpha, or teleport 
 * himself to a random location of the graph with probability 1-alpha */
double alpha = 0.8;
/* convergence_threshold controls the termination of the power iteration method 
 * for computing the stable distribution of the transition matrix made 
 * stochastic and primitive. */
double convergence_threshold = 1E-10;
size_t nb_res; /* nb_res is the number of resources */
/* res is an array of resources' names (i.e. URIs) */
char** res = NULL;
size_t nb_ter; /* nb_ter is the number of terms (or stems) */
/* ter is an array of terms' names (in fact the terms are stems) */
char** ter = NULL;
/* the user can provide indices of resources to be found in the res array that 
 * he knows to be important because they are semantically close to the query 
 * similarly for the terms */
size_t nb_res_close_to_query;
size_t* res_close_to_query = NULL;
size_t nb_ter_close_to_query;
size_t* ter_close_to_query = NULL;

/* resource-term matrix, "d" for "document" */
DMat d;
/* adjacency matrix */
DMat a;
/* transition probability matrix */
DMat p;
/* teleporting vector */
double* t;
/* stable distribution of the stochastic and primitive transition matrix */
double* pi;

int 
compare_idouble_value(void const * a, void const * b)
{
  IDouble const * da = (IDouble const *) a;
  IDouble const * db = (IDouble const *) b;
  return (da->value > db->value) - (da->value < db->value);
}

void 
idouble_sort(IDouble* values, long size)
{
  qsort(values, size, sizeof(IDouble), compare_idouble_value);
}

/* mk_teleport builds the teleportation  matrix
 *
 * A vector (i.e. array) t is used to store it (and not a matrix) since the 
 * target of the teleportation doesn't depend on the source.
 *
 * when there are no identified resources semantically close to the query, 
 * the teleportation matrix is uniform (each node has the same chance 
 * to be the target of the teleportation). 
 *
 * otherwise, the probability to teleport to a resource is set proportional to 
 * the maximum of the scalar products between this resource and the resources 
 * close to the query. In other words, the teleportation is not blind but 
 * is influenced by the knowledge we have of the user's information need as 
 * expressed by her query. */
void 
mk_teleport()
{
  double c1 = 0.0;
  double c2 = 0.0;
  double* v1;
  double* v2;
  double t_sum = 0.0;

  if (nb_res_close_to_query == 0)
  {
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      t[i] = 1.0 / nb_res;
    }
  } 
  else 
  {
    v1 = (double*) malloc(nb_ter * sizeof(double));
    check_mem(v1);
    v2 = (double*) malloc(nb_ter * sizeof(double));
    check_mem(v2);

    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      c1 = 0;
      c2 = 0;

      for (size_t j = 0 ; j < nb_ter ; ++j)
      {
        /* v1 is a line of the matrix d (i.e., the representation of a 
         * resource) */
        v1[j] = d->value[i][j];
      }

      for (size_t k = 0 ; k < nb_res_close_to_query ; ++k)
      {
        for (size_t j = 0 ; j < nb_ter ; ++j)
        {
          /* v2 is a line of the matrix d corresponding to a close-to-query 
           * resource */
          v2[j] = d->value[res_close_to_query[k]][j];
          /* c1 is the dot product between v1 and v2 */
          c1 += v1[j] * v2[j];
        }
        c1 = c1 / ( eucl_norm(nb_ter, v1) * eucl_norm(nb_ter, v2) );
        /* c2 stores the maximum of the cosinus between v1 and each of the 
         * resources close to the query */
        c2 = max(c1,c2);
      }

      /* Make a positive integer of c2 by keeping two digits 0.xx -> xx+1 */
      t[i] = round(c2*100) + 1;
      t_sum += t[i];
    }

    /* Make t a distribution (i.e., sum to 1) */
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      t[i] = t[i] / t_sum;
    }

    free(v1);
    free(v2);
  }
  return;
error:
  exit(EXIT_FAILURE);
}

/* mk_stochastic builds the stochastic matrix of transitions from resource to 
 * resource.
 *
 * for a resource with no neighbours, the target of the transition can be any 
 * other resource, the probability of the transition is then 
 * proportional to the lexical similarity between the resources as measured 
 * by the scalar product of their term vectors. 
 *
 * otherwise, the transition is towards a neighbour and is still 
 * proportional to the lexical similarity of the resources. */
void 
mk_stochastic()
{
  double* v1;
  double* v2;
  size_t* neighbours;
  double c;
  double sum;
  size_t nb_neighbours;

  v1 = (double*) malloc(nb_ter * sizeof(double));
  check_mem(v1);
  v2 = (double*) malloc(nb_ter * sizeof(double));
  check_mem(v2);
  neighbours = (size_t*) malloc(nb_res * sizeof(size_t));
  check_mem(neighbours);

  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    nb_neighbours = 0;
    /* v1 is the ith row of resource-term matrix d */
    for (size_t k = 0 ; k < nb_ter ; ++k) v1[k] = d->value[i][k];
    /* nb_neighbours of the ith resource
     * the neighbours' identifiers are stored in array neighbours */
    for (size_t j = 0 ; j < nb_res ; ++j)
    {
      /* initialize the transition probability matrix p with zeros */
      p->value[i][j] = 0;

      /* find the neighbours of the ith resource */
      if (a->value[i][j] > 1E-10)
      {
        neighbours[nb_neighbours] = j;
        nb_neighbours++;
      }
    }
    if (nb_neighbours == 0)
    {
      for (size_t j = 0 ; j < nb_res ; ++j)
      {
        c = 0.0;
        for (size_t k = 0 ; k < nb_ter ; ++k)
        {
          v2[k] = d->value[j][k];
          c += v1[k] * v2[k];
        }
        /* c is the cosine between resources i and j */
        c = c / ( eucl_norm(nb_ter, v1) * eucl_norm(nb_ter, v2) );
        /* Make a positive integer of c by keeping two digits 0.xx -> xx+1 
         * and store this value in transition probability matrix p
         * For now, p is not stochastic. The normalization will be done at the 
         * end of this function. */
        p->value[i][j] = round(c*100) + 1;
      }
    } 
    else /* resource i has at least one neighbour */
    {
      for (size_t j = 0 ; j < nb_neighbours ; ++j)
      {
        c = 0.0;
        for (size_t k = 0 ; k < nb_ter ; ++k)
        {
          v2[k] = d->value[neighbours[j]][k];
          c += v1[k] * v2[k];
        }
        c = c / ( eucl_norm(nb_ter, v1) * eucl_norm(nb_ter, v2) );
        p->value[i][neighbours[j]] = round(c*100) + 1;
      }
    }
  }

  /* normalization of the transition probability matrix p to make it 
   * stochastic */
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    sum = 0.0;
    for (size_t j = 0 ; j < nb_res ; ++j)
    {
      sum += p->value[i][j];
    }
    /* sum cannot be zero because 
     * either a resource i has no neighbours and each p(i,j) is positive,
     * or a resource i has at least one neighbour k, and p(i,k) is positive. */
    for (size_t j = 0 ; j < nb_res ; ++j)
    {
      p->value[i][j] = p->value[i][j] / sum;
    }
  }

  free(v1);
  free(v2);
  free(neighbours);
  return;
error:
  exit(EXIT_FAILURE);
}

/* mk_primitive makes the stochastic matrix p primitive by introducing 
 * a random teleportation factor.
 *
 * the importance of the teleportation factor is controlled by the alpha 
 * parameter. */
void 
mk_primitive()
{
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    for (size_t j = 0 ; j < nb_res ; ++j)
    {
      p->value[i][j] = alpha * p->value[i][j] + (1-alpha) * t[j];
    }
  }
}

/* power_it implements the power iteration method to compute the stable 
 * distribution of the stochastic and primitive matrix p 
 *
 * precision is controlled by the convergence_threshold parameter. */
void 
power_it()
{
  double* tmp;
  tmp = (double*) malloc(nb_res * sizeof(double));
  check_mem(tmp);
  for (;;)
  {
    for (size_t i = 0 ; i < nb_res ; ++i) tmp[i] = 0;
    for (size_t j = 0 ; j < nb_res ; ++j)
    {
      for (size_t i = 0 ; i < nb_res ; ++i)
      {
        /* product of the left row vector pi by the stochastic matrix p (pi'*p)
         * reminder: p  is the stochastic transition matrix,
         *           pi will converge to the eigen vector associated with 
         *              the eigen value 1. Since p is stochastic and primitive, 
         *              1 is indeed its greatest eigenvalue.
         *              pi/norm(pi) converges to the eigenvalue 1.
         *              If we consider p as the representation of a Markov 
         *              chain, pi converges to its stationary distribution. */
        tmp[j] += p->value[i][j] * pi[i];
      }
    }
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      pi[i] = tmp[i] / eucl_norm(nb_res, tmp);
    }
    /* terminate when pi/norm(pi) is close enough to the eigen value 1. */
    if (fabs(eucl_norm(nb_res,tmp) - 1) < convergence_threshold) break;
  }

  free(tmp);
  return;
error:
  exit(EXIT_FAILURE);
}

double 
eucl_norm(size_t n, double const * const v) 
{
  double max = 0.0;
  double sum = 0.0;
  
  for (size_t i = 0 ; i < n ; i++)
  {
    max = fmax(fabs(v[i]), max);
  }
  if (max < 1E-10) return 0.0;

  for (size_t i = 0 ; i < n ; i++)
  {
    sum += pow((v[i] / max), 2); 
  }
  return (max * sqrt(sum));
}

int 
main(int argc, char **argv)
{
  clock_t start_t, end_t, total_t;
  start_t = clock();

  FILE* fp;
  char* line = NULL;
  size_t len = 0;
  ssize_t read;
  char sep = ' ';
  char* seps = " \n";
  int c; /* option character for getopt */
  int errflg = 0;
  int show_clock = 0;
  extern char* optarg;
  extern int optind, optopt;
  size_t* m_line;
  IDouble* ret;

  while ((c = getopt(argc, argv, ":ca:t:")) != -1)
  {
    switch (c)
    {
      case 'c':
        show_clock++;
        break;
      case 'a':
        alpha = atof(optarg);
        break;
      case 't':
        convergence_threshold = atof(optarg);
        break;
      case ':': /* -a or -t without operand */
        fprintf(stderr, "Option -%c requires an operand\n", optopt);
        errflg++;
        break;
      case '?':
        fprintf(stderr, "Unrecognized option: -%c\n", optopt);
        errflg++;
    }
  }

  if (optind != argc - 1)
  {
    fprintf(stderr, "Wrong number of arguments\n");
    errflg++;
  }

  if (alpha == 0.0) /* mostly in case atof failed */
  {
    fprintf(stderr, "incorrect format for the alpha parameter (-a) \
        must be the string representation of a double object\n");
    errflg++;
  }

  if (convergence_threshold == 0.0) /* mostly in case atof failed */
  {
    fprintf(stderr, 
        "incorrect format for the convergence_threshold parameter (-t) \
        must be the string representation of a double object\n");
    errflg++;
  }

  if (errflg)
  {
    fprintf(stderr, "usage: ldpr [options] filepath\n\
-c: optional parameter, c stands for 'clock', when present the time\
spent by the algorithm is print out\n\n\
-a val: optional parameter, s stands for 'alpha', default value is 0.8,\
1-alpha represents the importance of the teleportation factor\n\n\
-t val: optional parameter, t stands for 'convergence_threshold', \
default value is 1E-10, it represents the precision of the power iteration\
method\n");
    exit(2);
  }

  fp = fopen(argv[optind], "r");
  if (fp == NULL)
  {
    fprintf(stderr, "ldpr: failed to open the resources-terms matrix file\n");
    exit(EXIT_FAILURE);
  }

  /* (1) list of resources */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    nb_res = count_char(line, sep) + 1;
    res = (char**) malloc(nb_res * sizeof(char*));
    check_mem(res);
    tokenize_str(line, seps, res);
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the resources\n");
    exit(EXIT_FAILURE);
  }

  /* (2) list of terms */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    nb_ter = count_char(line, sep) + 1;
    ter = (char**) malloc(nb_ter * sizeof(char*));
    check_mem(ter);
    tokenize_str(line, seps, ter);
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the terms\n");
    exit(EXIT_FAILURE);
  }

  /* (3) size of the resources-terms matrix */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    /* do nothing */
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the size of the matrix\n");
    exit(EXIT_FAILURE);
  }

  /* (4) resources-terms matrix line by line */
  d = svdNewDMat(nb_res, nb_ter);
  if (!d) 
  {
    fprintf(stderr, "ldpr: failed to allocate resource-term matrix\n");
    exit(EXIT_FAILURE);
  }

  for (int i = 0 ; i < nb_res ; ++i)
  {
    read = getline(&line, &len, fp);
    if (read == -1) 
    {
      fprintf(stderr, 
          "ldpr: error while reading a line of the resources-terms matrix\n");
      exit(EXIT_FAILURE);
    }

    m_line = (size_t*) malloc(nb_ter * sizeof(size_t));
    check_mem(m_line);
    tokenize_size_t(line, seps, m_line);

    for (int j = 0 ; j < nb_ter ; j++)
    {
      d->value[i][j] = (double)m_line[j];
    }

    free(m_line);
  }
  
  /* (5) resources close to the query */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    if (read == 1 && line[0] == 10) /* empty line: "\n" */
    {
      nb_res_close_to_query = 0;
    }
    else
    {
      nb_res_close_to_query = count_char(line,sep) + 1;
    }
    res_close_to_query = (size_t*) malloc(nb_res_close_to_query * 
                                          sizeof(size_t));
    check_mem(res_close_to_query);
    tokenize_size_t(line, seps, res_close_to_query);
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the resources close to the query\n");
    exit(EXIT_FAILURE);
  }
  
  /* (6) terms close to the query */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    if (read == 1 && line[0] == 10) /* empty line: "\n" */
    {
      nb_ter_close_to_query = 0;
    }
    else
    {
      nb_ter_close_to_query = count_char(line,sep) + 1;
    }
    ter_close_to_query = (size_t*) malloc(nb_ter_close_to_query * 
                                          sizeof(size_t));
    check_mem(ter_close_to_query);
    tokenize_size_t(line, seps, ter_close_to_query);
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the terms close to the query\n");
    exit(EXIT_FAILURE);
  } 

  /* (7) adjacency matrix line by line */
  a = svdNewDMat(nb_res, nb_res);
  if (!a) 
  {
    fprintf(stderr, "ldpr: failed to allocate adjacency matrix\n");
    exit(EXIT_FAILURE);
  }

  for (int i = 0 ; i < nb_res ; ++i)
  {
    read = getline(&line, &len, fp);
    if (read == -1) 
    {
      fprintf(stderr, 
              "ldpr: error while reading a line of the adjacency matrix\n");
      exit(EXIT_FAILURE);
    }

    m_line = (size_t*) malloc(nb_res * sizeof(size_t));
    check_mem(m_line);
    tokenize_size_t(line, seps, m_line);

    for (int j = 0 ; j < nb_res ; j++)
    {
      a->value[i][j] = (double)m_line[j];
    }

    free(m_line);
  }

  fclose(fp);

  t = (double*) malloc(nb_res * sizeof(double));
  check_mem(t);
  pi = (double*) malloc(nb_res * sizeof(double));
  check_mem(pi);
  for (size_t i = 0 ; i < nb_res ; ++i) pi[i] = 1.0 / nb_res;
  p = svdNewDMat(nb_res, nb_res);

  mk_teleport();
  mk_stochastic();
  mk_primitive();
  power_it();

  if (show_clock)
  {
    end_t = clock();
    total_t = end_t - start_t;
    printf("%lu\n", total_t);
    /*printf("CLOCKS_PER_SEC = %ld\n", CLOCKS_PER_SEC);*/
  }

  /* print result */
  ret = (IDouble*) malloc(nb_res * sizeof(IDouble));
  check_mem(ret);

  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    ret[i].index = i;
    ret[i].value = pi[i];
  }

  idouble_sort(ret, nb_res);

  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    printf("%s %f\n", res[ret[i].index], ret[i].value);
  }

  exit(EXIT_SUCCESS);
error:
  exit(EXIT_FAILURE);
}
