/*
 * Copyright 2014 Pierre-Edouard Portier, Mazen Alsarem
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <search.h>
#include <time.h>
#include <unistd.h>
#include "dbg.h"
#include "svdlib.h"
#include "str.h"
#include "ldpr.h"

int 
compare_idouble_value(void const * a, void const * b)
{
  IDouble const * da = (IDouble const *) a;
  IDouble const * db = (IDouble const *) b;
  return (da->value > db->value) - (da->value < db->value);
}

void 
idouble_sort(IDouble* values, long size)
{
  qsort(values, size, sizeof(IDouble), compare_idouble_value);
}

double 
eucl_norm(size_t n, double const * const v) 
{
  double max = 0.0;
  double sum = 0.0;
  
  for (size_t i = 0 ; i < n ; i++)
  {
    max = fmax(fabs(v[i]), max);
  }
  if (max < 1E-10) return 0.0;

  for (size_t i = 0 ; i < n ; i++)
  {
    sum += pow((v[i] / max), 2); 
  }
  return (max * sqrt(sum));
}

/* eucl_norm_cols returns an array of the euclidean norms of the columns of m*/
double* 
eucl_norm_cols (DMat const m) 
{
  /* r is the returned value */
  double* r;
  /* vs is an array that represents a dense view of the matrix m,
   * columns by columns, in a contiguous zone of the memory. */
  double* vs;
  size_t vs_size = m->rows * m->cols;

  r = (double*) malloc(sizeof(double) * m->cols);
  check_mem(r);
  vs = (double*) malloc(sizeof(double) * vs_size);
  check_mem(vs);

  for (size_t i = 0 ; i < m->rows ; i++) 
  {
    for (size_t j = 0 ; j < m->cols ; j++) 
    {
      vs[j * m->rows + i] = m->value[i][j];
    }
  }

  for (size_t i = 0 ; i < m->cols ; i++) 
  {
    r[i] = eucl_norm(m->rows, &vs[i * m->rows]);
  }

  free(vs);
  return r;

error:
  exit(EXIT_FAILURE);
}

/* Multiply each row k of m by s[k].
 * The size of s must be the number of rows of m.
 * This operation is applied on the results of a SVD. Thus, the parameter m 
 * is a dense matrix. */
DMat 
scale_rows (DMat const m, double const * const s)
{
  DMat d = svdNewDMat(m->rows, m->cols);
  for (size_t i = 0 ; i < m->rows ; i++)
  {
    for (size_t j = 0 ; j < m->cols ; j++)
    {
      d->value[i][j] = m->value[i][j] * s[i];
    }
  }
  return d;
}

int 
main(int argc, char **argv)
{
  clock_t start_t, end_t, total_t;
  start_t = clock();

  FILE* fp;
  char* line = NULL;
  char* line_nz = NULL;
  char* line_rowind = NULL;
  char* line_pointr = NULL;
  size_t len = 0;
  ssize_t read;
  char sep = ' ';
  char* seps = " \n";
  int c; /* option character for getopt */
  int errflg = 0;
  int show_clock = 0;

  /* the importance of the teleportation factor is controlled by the alpha 
   * parameter. At each step, the random walker choose to either reach one of 
   * the neighbour positions in the graph with probability alpha, or teleport 
   * himself to a random location of the graph with probability 1-alpha */
  double alpha = 0.8;
  /* convergence_threshold controls the termination of the power iteration method 
   * for computing the stable distribution of the transition matrix made 
   * stochastic and primitive. */
  double convergence_threshold = 1E-10;
  size_t nb_res; /* nb_res is the number of resources */
  /* res is an array of resources' names (i.e. URIs) */
  char** res = NULL;
  size_t nb_ter; /* nb_ter is the number of terms (or stems) */
  size_t rm_nb_nz; /* nb of non zero (nz) values of the 
                      resources-terms (rm) matrix */
  size_t am_nb_nz; /* nb of non zero (nz) values of the 
                      adjacency (am) matrix */
  /* the user can provide indices of resources to be found in the res array that 
   * he knows to be important because they are semantically close to the query 
   * similarly for the terms */
  size_t nb_res_close_to_query;
  size_t* res_close_to_query = NULL;
  size_t nb_ter_close_to_query;
  size_t* ter_close_to_query = NULL;
  /* sparse resources-terms matrix */
  SMat rm;
  /* sparse adjacency matrix */
  SMat am;

  int i;
  int rank; /* Temporally store the rank of a doc. Used for computing the 
               hitscore of a resource. */
  int nbdocs; /* Total number of documents. */
  int total_hitscore = 0;
  size_t dn_size = 0; /* number of dangling nodes */
  size_t* dn; /* indices of the dangling nodes */
  extern char* optarg;
  extern int optind, optopt;
  IDouble* ret;
  SMat amt; /* transpose of am */
  SMat rmt = NULL; /* transpose of rm */
  SMat modified_rm = NULL;
  /* Stable distribution of the stochastic and primitive transition matrix. */
  double* pi;
  /* Component of pi. */
  double* pi_c1;
  double pi_c2;
  double* teleport;
  double residual;
  SVDRec svd1 = NULL;
  SVDRec svd2 = NULL;
  DMat sut;
  double* sm1;
  /* matrix S^-1 * Vt used to transform a resource vector (a line of A) into
   * a topic vector (line of U). Indeed:
   * A = U S Vt
   * A V S^-1 = U
   * Ut = S^-1 Vt At
   * Thus, given a vector x in the original space,
   * S^-1 Vt xt is a representation of x in the reduced space. */
  DMat sm1vt = NULL;
  /* qk is a vector (array) of length nb_dim
   * It lies in the reduced space and is computed so as to point in a 
   * "good" direction, i.e. a direction that should be close to the one 
   * corresponding to the information need (given the query, the RDF graph, and 
   * the text associated to each concept) */
  double* qk = NULL;
  double* res_norms = NULL;
  double* prev_res_norms = NULL;
  int is_a_res_close_to_query;
  IDouble* res_deltas_s = NULL;
  size_t nb_top_res;
  double top_threshold = 0.01;
  int nb_dim = 10;
  double stress_factor = 1000.0;
  double* u = NULL; /* u and v are vectors of resources in the reduced space */
  double* v = NULL;
  /* s_product stores the scalar product of vectors u and v */
  double s_product;
  /* cosine stores the cosine corresponding to s_product */
  double cosine;
  double sum_cosine;
  double sum_resscore;
  size_t rid;
  double* hitscore = NULL; /* hitscore by increasing order resources indices */
  double* resscore_a = NULL; /* resscore by increasing order resources indices */
  double* resscore_b = NULL; /* resscore by increasing order resources indices */
  double* terscore = NULL; /* hitscore by increasing order resources indices */
  
  /****************************************************************************/
  /* Parse command line options. */
  /****************************************************************************/
  
  while ((c = getopt(argc, argv, ":ca:t:")) != -1)
  {
    switch (c)
    {
      case 'c':
        show_clock++;
        break;
      case 'a':
        alpha = atof(optarg);
        break;
      case 't':
        convergence_threshold = atof(optarg);
        break;
      case ':': /* -a or -t without operand */
        fprintf(stderr, "Option -%c requires an operand\n", optopt);
        errflg++;
        break;
      case '?':
        fprintf(stderr, "Unrecognized option: -%c\n", optopt);
        errflg++;
    }
  }

  if (optind != argc - 1)
  {
    fprintf(stderr, "Wrong number of arguments\n");
    errflg++;
  }

  if (alpha == 0.0) /* most likely in case atof failed */
  {
    fprintf(stderr, "incorrect format for the alpha parameter (-a) \
        must be the string representation of a double object\n");
    errflg++;
  }

  if (convergence_threshold == 0.0) /* most likely in case atof failed */
  {
    fprintf(stderr, 
        "incorrect format for the convergence_threshold parameter (-t) \
        must be the string representation of a double object\n");
    errflg++;
  }

  if (errflg)
  {
    fprintf(stderr, "usage: ldpr [options] filepath\n\
-c: optional parameter, c stands for 'clock', when present the time\
spent by the algorithm is print out\n\n\
-a val: optional parameter, s stands for 'alpha', default value is 0.8,\
1-alpha represents the importance of the teleportation factor\n\n\
-t val: optional parameter, t stands for 'convergence_threshold', \
default value is 1E-10, it represents the precision of the power iteration\
method\n");
    exit(2);
  }
  
  /****************************************************************************/
  /* Parse input file. */
  /****************************************************************************/
  
  fp = fopen(argv[optind], "r");
  if (fp == NULL)
  {
    fprintf(stderr, "ldpr: failed to open the input file\n");
    exit(EXIT_FAILURE);
  }

  /* (1) list of resources */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    nb_res = count_char(line, sep) + 1;
    res = (char**) malloc(nb_res * sizeof(char*));
    check_mem(res);
    tokenize_str(line, seps, res);
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the resources\n");
    exit(EXIT_FAILURE);
  }

  if (nb_res == 0)
  {
    fprintf(stderr, "ldpr: there are no resources\n");
    exit(EXIT_FAILURE);
  }

  /* (2) binary vectors to compute the hitscore of each resource */
  hitscore = (double*) calloc(nb_res, sizeof(double));
  check_mem(hitscore);

  if ((read = getline(&line, &len, fp)) != -1)
  {
    nbdocs = 0;
    while(line[nbdocs] != ' ')
    {
      nbdocs++;
    }
    i = 0;
    for (int j = 0 ; j < nb_res ; ++j)
    {
      rank = 1;
      while(line[i] != ' ')
      {
        if (line[i] == '1')
        {
          hitscore[j] += (nbdocs + 1) - rank;
        }
        i++; rank++;
      }
      i++;
      total_hitscore += hitscore[j];
    }
    for (int i = 0 ; i < nb_res ; ++i)
    {
      hitscore[i] = (double) hitscore[i] / total_hitscore;
    }
  }
  else 
  {
    fprintf(stderr, "ldpr: failed to read the binary vectors used to compute \
      the resources' hitscores\n");
    exit(EXIT_FAILURE);
  }

  /* (3) non-zero (nz) values of the resources-terms matrix (rm) \
  column by column */
  if ((read = getline(&line_nz, &len, fp)) != -1)
  {
    if (read == 1 && line_nz[0] == 10) /* empty line: "\n" */
    {
      rm_nb_nz = 0;
    }
    else
    {
      rm_nb_nz = count_char(line_nz, sep) + 1;
    }
  }
  else 
  {
    fprintf(stderr, "ldpr: failed to read the nz values of the resources-terms \
      matrix\n");
    exit(EXIT_FAILURE);
  }

  /* (4) row index of the nz values of rm */
  if ((read = getline(&line_rowind, &len, fp)) != -1)
  {
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the rowind of the nz values of rm\n");
    exit(EXIT_FAILURE);
  }

  /* (5) for each col of rm, index in rm_nz of the first nz entry*/
  if ((read = getline(&line_pointr, &len, fp)) != -1)
  {
    nb_ter = count_char(line_pointr, sep) + 1;
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the pointr struct for rm\n");
    exit(EXIT_FAILURE);
  }

  /* Build the sparse resources-terms (rm) matrix 
     svdNewSMat expects values of type int for its three parameters 
     (nbrows, nbcols, and nbvals). But nb_res, nb_ter, and rm_nb_nz are of type 
     size_t (which is a good thing...). Thus, the arguments are explicitly cast 
     and in very rare cases, it could be the cause of bugs.*/
  rm = svdNewSMat((int) nb_res, (int) nb_ter, (int) rm_nb_nz);
  if (!rm) 
  {
    fprintf(stderr, "ldpr: failed to allocate sparse resources-terms matrix\n");
    exit(EXIT_FAILURE);
  }

  tokenize_double(line_nz, seps, rm->value);
  tokenize_long(line_rowind, seps, rm->rowind);
  tokenize_long(line_pointr, seps, rm->pointr);
  
  /* (6) resources close to the query */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    if (read == 1 && line[0] == 10) /* empty line: "\n" */
    {
      nb_res_close_to_query = 0;
    }
    else
    {
      nb_res_close_to_query = count_char(line,sep) + 1;
    }
    res_close_to_query = (size_t*) malloc(nb_res_close_to_query * 
                                          sizeof(size_t));
    check_mem(res_close_to_query);
    tokenize_size_t(line, seps, res_close_to_query);
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the resources close to the query\n");
    exit(EXIT_FAILURE);
  }
  
  /* (7) terms close to the query */
  if ((read = getline(&line, &len, fp)) != -1)
  {
    if (read == 1 && line[0] == 10) /* empty line: "\n" */
    {
      nb_ter_close_to_query = 0;
    }
    else
    {
      nb_ter_close_to_query = count_char(line,sep) + 1;
    }
    ter_close_to_query = (size_t*) malloc(nb_ter_close_to_query * 
                                          sizeof(size_t));
    check_mem(ter_close_to_query);
    tokenize_size_t(line, seps, ter_close_to_query);
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the terms close to the query\n");
    exit(EXIT_FAILURE);
  }

/* (8) non-zero (nz) values of the adjacency matrix (am) \
  column by column */
  if ((read = getline(&line_nz, &len, fp)) != -1)
  {
    if (read == 1 && line_nz[0] == 10) /* empty line: "\n" */
    {
      am_nb_nz = 0;
    }
    else
    {
      am_nb_nz = count_char(line_nz, sep) + 1;
    }
  } 
  else
  {
    fprintf(stderr, "ldpr: failed to read the nz values of the adjacency \
      matrix\n");
    exit(EXIT_FAILURE);
  }

  /* (9) row index of the nz values of am */
  if ((read = getline(&line_rowind, &len, fp)) != -1)
  {
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the rowind of the nz values of am\n");
    exit(EXIT_FAILURE);
  }

  /* (10) for each col of am, index in am_nz of the first nz entry*/
  if ((read = getline(&line_pointr, &len, fp)) != -1)
  {
  } 
  else 
  {
    fprintf(stderr, "ldpr: failed to read the pointr struct for am\n");
    exit(EXIT_FAILURE);
  }

  fclose(fp);

  /* Build the sparse adjacency (am) matrix 
     svdNewSMat expects values of type int for its three parameters 
     (nbrows, nbcols, and nbvals). But nb_res, nb_ter, and rm_nb_nz are of type 
     size_t (which is a good thing...). Thus, the arguments are explicitly cast 
     and in very rare cases, it could be the cause of bugs.*/
  am = svdNewSMat((int) nb_res, (int) nb_res, (int) am_nb_nz);
  if (!am) 
  {
    fprintf(stderr, "ldpr: failed to allocate sparse adjacency matrix\n");
    exit(EXIT_FAILURE);
  }

  tokenize_double(line_nz, seps, am->value);
  tokenize_long(line_rowind, seps, am->rowind);
  tokenize_long(line_pointr, seps, am->pointr);

  free(line);
  free(line_nz);
  free(line_rowind);
  free(line_pointr);

  /****************************************************************************/
  /* (A) Find important resources.                                            */
  /****************************************************************************/

  /****************************************************************************/
  /* (A.1) Use the resources annotated in the query or its wordnet extension. */
  /****************************************************************************/
  if(0){
  svd1 = svdLAS2A(rm, nb_dim);
  /* in the new reduced space, 
   * scale the resources by the scaling factors of Sigma */
  sut = scale_rows(svd1->Ut, svd1->S);

  sm1 = (double*) malloc(sizeof(double) * nb_dim);
  check_mem(sm1);
  for (size_t j = 0 ; j < nb_dim ; ++j)
  {
    sm1[j] = 1.0 / (svd1->S)[j];
  }
  /* S^-1 * Vt the operator used to convert a resource represented as a 
   * combination of stems (line of A) into a resource represented as a 
   * combination of topics (line of U) */
  sm1vt = scale_rows(svd1->Vt, sm1);

  nb_dim = min(nb_dim, nb_res);
  if (nb_res_close_to_query > 0)
  {
  /****************************************************************************/
  /* (A.1.1) Find important resources related to the resources annotated in the
  query.                                                                      */
  /****************************************************************************/
    prev_res_norms = eucl_norm_cols(sut);
    rmt = svdTransposeS(rm);

    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      is_a_res_close_to_query = 0;
      for(size_t j = 0 ; j < nb_res_close_to_query ; ++j)
      {
        if (res_close_to_query[j] == i)
        {
          is_a_res_close_to_query = 1;
          break;
        }
      }
      if (is_a_res_close_to_query)
      {
        for (long j = rmt->pointr[i] ; j < rmt->pointr[i+1] ; ++j)
        {
          rmt->value[j] *= stress_factor;
        }
      }
    }

    modified_rm = svdTransposeS(rmt);
    svd2 = svdLAS2A(modified_rm, nb_dim);
    sut = scale_rows(svd2->Ut, svd2->S);
    res_norms = eucl_norm_cols(sut);

    res_deltas_s = (IDouble*) malloc(sizeof(IDouble) * nb_res);
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      res_deltas_s[i].index = i;
      res_deltas_s[i].value = fabs(res_norms[i] - prev_res_norms[i]);
    }
  
  /****************************************************************************/
  /* (A.1.2.a) Build a probability vector of the importance of the resources. 
  The probability of a resource is proportional to the distance it moved 
  from the origin during (A.1.1.). */
  /****************************************************************************/
    resscore_b = (double*) malloc(sizeof(double) * nb_res);
    check_mem(resscore_b);
    sum_resscore = 0.0;
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      resscore_b[i] = res_deltas_s[i].value;
      sum_resscore += resscore_b[i];
    }
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      resscore_b[i] = (double) resscore_b[i] / sum_resscore;
    }

  /****************************************************************************/
  /* (A.1.2.b) Build a probability vector of the importance of the resources. 
  The probability of a resource is proportional to the (mean / max) of the 
  cosine similarity in the reduced space between this resource and each of the 
  important resources of (A.1.1.). */
  /****************************************************************************/
    idouble_sort(res_deltas_s, nb_res);
    nb_top_res = (size_t) ceil((double) top_threshold * nb_res);
    u = (double*) malloc(sizeof(double) * nb_dim);
    check_mem(u);
    v = (double*) malloc(sizeof(double) * nb_dim);
    check_mem(v);
    resscore_a = (double*) malloc(sizeof(double) * nb_res);
    check_mem(resscore_a);
    sum_cosine = 0.0;
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      cosine = 0.0;
      for (size_t j = 1 ; j <= nb_top_res ; ++j)
      {
        rid = res_deltas_s[nb_res - j].index;

        s_product = 0.0;
        for (size_t k = 0 ; k < nb_dim ; ++k)
        {
          u[k] = (svd1->Ut)->value[k][rid];
          v[k] = (svd1->Ut)->value[k][i];
          s_product += u[k] * v[k];
        }
        cosine = max(cosine, s_product / ( eucl_norm(nb_dim,u) * eucl_norm(nb_dim,v) ));
      }
      resscore_a[i] = cosine;
      sum_cosine += cosine;
    }
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      resscore_a[i] = (double) resscore_a[i] / sum_cosine;
    }
  }

  /****************************************************************************/
  /* (A.2) Use the terms present in the query or in its wordnet extension. */
  /****************************************************************************/

  /****************************************************************************/
  /* (A.2.1) With these terms, build a binary vector q and map it to the reduced 
  space. */
  /****************************************************************************/
  if (0){
  qk = (double*) malloc(sizeof(double) * nb_dim);
  check_mem(qk);
  for (size_t i = 0 ; i < nb_dim ; ++i)
  {
    qk[i] = 0.0;
    for (size_t k = 0 ; k < nb_ter_close_to_query ; ++k)
    {
      qk[i] += sm1vt->value[i][ter_close_to_query[k]];
    }
  }

  /****************************************************************************/
  /* (A.2.2) Build a probability vector of the importance of the resources.
  The probability of a resource is proportional to its cosine similarity with 
  the image of q in the reduced space. */
  /****************************************************************************/
  terscore = (double*) malloc(sizeof(double) * nb_res);
  check_mem(terscore);
  sum_cosine = 0.0;
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    cosine = 0.0;

    s_product = 0.0;
    for (size_t k = 0 ; k < nb_dim ; ++k)
    {
      v[k] = (svd1->Ut)->value[k][i];
      s_product += qk[k] * v[k];
    }
    cosine = s_product / ( eucl_norm(nb_dim,qk) * eucl_norm(nb_dim,v) );

    terscore[i] = cosine;
    sum_cosine += cosine;
  }
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    terscore[i] = (double) terscore[i] / sum_cosine;
  }
  }/* if(0) */

  /* Free a few objects */
  svdFreeSVDRec(svd2);
  svdFreeSVDRec(svd1);
  svdFreeDMat(sut);
  svdFreeSMat(rmt);
  svdFreeSMat(modified_rm);
  free(sm1);
  free(sm1vt);
  free(u);
  free(v);
  free(qk);
  free(res_deltas_s);
  free(res_norms);
  free(prev_res_norms);
  } /* if(0) */

  /****************************************************************************/
  /* (A.3) Build a probability vector of the importance of the resources based 
  on the hitscore of each resource. 
  Already done above.*/
  /****************************************************************************/

  /****************************************************************************/
  /* (B) With belief aggregation, build a teleportation vector. 
  Use linear opinion pool with equal weight assigned to the three probability 
  vectors. */
  /****************************************************************************/
  if (0)
  {
    teleport = (double*) malloc(nb_res * sizeof(double));
    for (size_t i = 0 ; i < nb_res ; ++i)
    {
      teleport[i] = (double) (resscore_a[i] + resscore_b[i] + hitscore[i]) / 3.0;
    }
  }
  else
  {
    teleport = hitscore;
  }
  /*teleport = (double*) malloc(nb_res * sizeof(double));
  check_mem(teleport);
  for (long i = 0 ; i < nb_res ; ++i) teleport[i] = 1.0 / nb_res;*/

  /****************************************************************************/
  /* (C) Compute the transition probability matrix. */
  /****************************************************************************/
  /* The Compressed Column Storage (CCS) of am is not adequate to manipulate the 
  rows of am. Thus, we transpose am into amt. */
  amt = svdTransposeS(am);
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    double sum_am_row_i = 0;
    for (long j = amt->pointr[i] ; j < amt->pointr[i+1] ; ++j)
    {
      sum_am_row_i += amt->value[j];
    }
    for (long j = amt->pointr[i] ; j < amt->pointr[i+1] ; ++j)
    {
      amt->value[j] = amt->value[j] / sum_am_row_i;
    }
  }

  /****************************************************************************/
  /* (D) Add virtual outgoing links to the 5% dangling nodes that are the 
  closest to important resources. */
  /****************************************************************************/

  /****************************************************************************/
  /* (D.1) Identify the dangling nodes. */
  /****************************************************************************/
  /* We are looking for empty rows of am. 
  The Compressed Column Storage (CCS) we use to manipulate am is useful to find 
  empty columns (with am->pointr) but not to find empty rows. Thus, we transpose 
  am into amt. */
  dn = (size_t*) malloc(nb_res * sizeof(size_t));
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    if ((amt->pointr[i+1] - amt->pointr[i]) == 0)
    {
      dn[dn_size++] = i;
    }
  }

  /****************************************************************************/
  /* (D.2) Rank the dangling nodes by computing their cosine similarities in the 
  reduced space with the first 5% of the resources when ordered by importance.*/
  /****************************************************************************/

  /****************************************************************************/
  /* (D.3) Update the transition probability matrix by replacing the first 5% of 
  the ranked dangling nodes with probability vectors whose non-zero values 
  represent outgoing links towards the most important nodes (the first 5% of the 
  nodes when ranked by importance). 
  These values are proportional to the cosine similarity in the reduced space 
  between the previously dangling node and the most important resources. */
  /****************************************************************************/

  /****************************************************************************/
  /* (D.4) Identify the indices of the remaining dangling nodes. */
  /****************************************************************************/

  /****************************************************************************/
  /* (E) Compute the final ranking by the power iteration method. */
  /****************************************************************************/

  /* Initialize pi. */
  pi = (double*) malloc(nb_res * sizeof(double));
  check_mem(pi);
  for (long i = 0 ; i < nb_res ; ++i) pi[i] = 1.0 / nb_res;

  pi_c1 = (double*) malloc(nb_res * sizeof(double));
  check_mem(pi_c1);
  for (;;)
  {
    for (long i = 0 ; i < nb_res ; ++i)
    {
      pi_c1[i]  = 0;
    }
    pi_c2 = 0;
    
    /* Compute pi_c1, the component that depends on the transition probability 
    matrix. We benefit from the sparisity of the matrix. */
    for (long j = 0 ; j < nb_res ; ++j)
    {
      for (long i = amt->pointr[j] ; i < amt->pointr[j+1] ; ++i)
      {
        pi_c1[amt->rowind[i]] += alpha * amt->value[i] * pi[j];
      }
    }

    for (long i = 0 ; i < dn_size ; ++i)
    {
      pi_c2 += alpha * pi[dn[i]];
    }

    pi_c2 += 1 - alpha;

    for (long i = 0 ; i < nb_res ; ++i)
    {
      pi_c1[i] += pi_c2 * teleport[i];
    } 

    residual = 0;
    for (long i = 0 ; i < nb_res ; ++i)
    {
      residual += fabs(pi_c1[i] - pi[i]);
    }
    for (long i = 0 ; i < nb_res ; ++i)
    {
      pi[i] = pi_c1[i];
    }
    if (residual < convergence_threshold) break;
  }

  free(pi_c1);

  /****************************************************************************/
  /* (F) Print result. */
  /****************************************************************************/
  if (show_clock)
  {
    end_t = clock();
    total_t = end_t - start_t;
    printf("%lu\n", total_t);
    /*printf("CLOCKS_PER_SEC = %ld\n", CLOCKS_PER_SEC);*/
  }

  ret = (IDouble*) malloc(nb_res * sizeof(IDouble));
  check_mem(ret);
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    ret[i].index = i;
    ret[i].value = pi[i];
  }
  idouble_sort(ret, nb_res);
  for (size_t i = 0 ; i < nb_res ; ++i)
  {
    printf("%s %f\n", res[ret[i].index], ret[i].value);
  }

  exit(EXIT_SUCCESS);
error:
  exit(EXIT_FAILURE);
}
